from feynamp.momentum import insert_momentum
from feynamp.util import find_particle_in_model


def get_leg_math_string(leg, fd, model):
    return get_leg_math(leg, fd, model)


def get_leg_math(fd, leg, model):  # epsilons or u/v optionally also barred
    p = find_leg_in_model(fd, leg, model)
    if p.particle.momentum is None or p.particle.momentum.name is None:
        raise ValueError("Momentum not set for particle")
    ret = ""
    mom = insert_momentum(p.particle.momentum.name)
    # give particles color vectors to sum over them in the end (or better average)
    # TODO this could be also done as incoming vs outcoming
    if p.color == 8:
        # if particle is a gluon give it a adjoint color function
        ret += f"VA(Glu{p.particle.id},{mom})*"
    if p.color == 3 or p.color == -3:
        # if particle is a quark give it a fundamental color function
        ret += f"VC(Col{p.particle.id},{mom})*"

    if p.spin == 3:
        if leg.is_incoming():
            ret += f"eps(Mu{p.particle.id},Pol{p.particle.id},{mom})"
        else:
            ret += f"eps_star(Mu{p.particle.id},Pol{p.particle.id},{mom})"
    if p.spin == 2:
        if not p.particle.is_anti():
            if leg.is_incoming():
                ret += f"u(Spin{p.particle.id},{mom})"
            else:
                ret += f"u_bar(Spin{p.particle.id},{mom})"
        else:
            if leg.is_incoming():
                ret += f"v(Spin{p.particle.id},{mom})"
            else:
                ret += f"v_bar(Spin{p.particle.id},{mom})"
    return ret


def find_leg_in_model(fd, leg, model):  # find leg in model
    assert leg in fd.legs
    return find_particle_in_model(leg, model)



import re
from typing import List

import numpy as np
from feynml.connector import Connector
from feynml.id import generate_new_id
from feynml.leg import Leg
from feynml.propagator import Propagator
from feynml.vertex import Vertex

from feynamp.momentum import insert_momentum
from feynamp.util import safe_index_replace


def insert_color_types(s: str):
    """ """
    # We use the non greedy .*? to match multiple occurances individually
    s = re.sub(r"T\((.*?),(.*?),(.*?)\)", r"T(Glu\1,Col\2,Col\3)", s)
    s = re.sub(r"f\((.*?),(.*?),(.*?)\)", r"f(Glu\1,Glu\2,Glu\3)", s)
    s = re.sub(r"Identity\((.*?),(.*?)\)", r"Identity(Col\1,Col\2)", s)
    return s


def insert_lorentz_types(s: str, connections: List[Connector], vertex: Vertex):
    # We use the non greedy .*? to match multiple occurances individually
    s = re.sub(r"Gamma\((.*?),(.*?),(.*?)\)", r"Gamma(Mu\1,Spin\2,Spin\3)", s)
    s = re.sub(r"ProjP\((.*?),(.*?)\)", r"ProjP(Spin\1,Spin\2)", s)
    s = re.sub(r"ProjM\((.*?),(.*?)\)", r"ProjM(Spin\1,Spin\2)", s)
    s = re.sub(r"Metric\((.*?),(.*?)\)", r"Metric(Mu\1,Mu\2)", s)
    # use insert_momentum to replace the second argument to P
    matches = re.findall(r"P\((.*?),(.*?)\)", s)
    for g in matches:
        # find connection with g[1] id in connections
        repl = None
        for c in connections:
            if c.id == g[1] or "In" + str(c.id) == g[1] or "Out" + str(c.id) == g[1]:
                # TODO: is this correct?
                if c.goes_into(vertex):
                    repl = "P(Mu" + g[0] + "," + insert_momentum(c.momentum.name) + ")"
                    break
                else:
                    repl = (
                        "(-P(Mu" + g[0] + "," + insert_momentum(c.momentum.name) + "))"
                    )
                    break
        if repl is None:
            raise Exception(
                f"Connection with id {g[1]} not found in connections {connections}"
            )
        s = s.replace("P(" + g[0] + "," + g[1] + ")", repl)
    return s


# def insert_index_types(s):
#    s = insert_color_types(s)
#    s = insert_lorentz_types(s)
#    return s


def get_vertex_math_string(fd, vertex, model):
    vv = get_vertex_math(fd, vertex, model)
    s = ""
    for v in vv:
        s += f"({v[0]})*({v[1]})*({v[2]}) + "
    return s[:-3]


def get_index_prefix(connection: Connector, vertex):
    if connection.goes_into(vertex):
        return "In"
    else:
        return "Out"


def get_vertex_math(fd, vertex, model, typed=True):  # TODO subst negative indices
    if not typed:
        raise NotImplementedError("Only typed vertices are supported")
    vv = fd.get_connections(vertex)
    v = find_vertex_in_model(fd, vertex, model)
    if v is None:
        raise Exception(f"Vertex {vertex} not found in model")
        # return None
    assert len(v.color) == len(v.lorentz)
    cret = []
    lret = []
    for j in range(len(v.color)):
        col = v.color[j]
        nid = generate_new_id()
        col = safe_index_replace(col, str(-1), str(nid))
        for i, vv in enumerate(v.particles):
            if isinstance(v.connections[i], Leg):
                col = safe_index_replace(col, str(i + 1), str(v.connections[i].id))
            elif isinstance(v.connections[i], Propagator):
                col = safe_index_replace(
                    col,
                    str(i + 1),
                    get_index_prefix(v.connections[i], vertex)
                    + str(v.connections[i].id),
                )
            else:
                raise Exception(
                    f"Connection {v.connections[i]} not a leg or propagator"
                )
        if typed:
            col = insert_color_types(col)
        cret.append(col)
    for k in range(len(v.lorentz)):
        lor = v.lorentz[j].structure
        nid = generate_new_id()
        lor = safe_index_replace(lor, str(-1), str(nid))
        for i, vv in enumerate(v.particles):
            if isinstance(v.connections[i], Leg):
                lor = safe_index_replace(lor, str(i + 1), str(v.connections[i].id))
            elif isinstance(v.connections[i], Propagator):
                lor = safe_index_replace(
                    lor,
                    str(i + 1),
                    get_index_prefix(v.connections[i], vertex)
                    + str(v.connections[i].id),
                )
            else:
                raise Exception(
                    f"Connection {v.connections[i]} not a leg or propagator"
                )
        if typed:
            lor = insert_lorentz_types(lor, v.connections, vertex)
        lret.append(lor)
    ret = []
    for k, v in v.couplings.items():
        ret.append((v.value, cret[k[0]], lret[k[1]]))
    return ret


def find_vertex_in_model(fd, vertex, model):
    """
    Finds the model vertex corresponding to the given FeynmanDiagram vertex

    Note: Sorting is to check for the correct particles in a vertex given they can be in any order and have duplicates
    """
    assert vertex in fd.vertices
    cons = np.array(fd.get_connections(vertex))
    # debug(f"{cons=}")
    pdg_ids_list = []

    # correct for incoming vs outgoing fermion struct
    for c in cons:
        p = c.pdgid
        if c.is_any_fermion():
            if c.goes_into(vertex):
                p = -p
        pdg_ids_list += [p]
    pdg_ids_array = np.array(pdg_ids_list)

    sort_mask = np.argsort(pdg_ids_array)
    particles = pdg_ids_array[sort_mask]
    scons = cons[sort_mask]
    # debug(f"{scons=}")
    ret = None
    for v in model.vertices:
        if len(v.particles) != len(particles):
            continue
        model_particle_ids = np.array([p.pdg_code for p in v.particles])
        model_sort_mask = np.argsort(model_particle_ids)
        # By sorting based on the indices we reproduce the order of the particles in the vertex
        inverted_model_sort_mask = np.argsort(model_sort_mask)
        sorted_model_particle_ids = model_particle_ids[model_sort_mask]
        if np.array_equal(sorted_model_particle_ids, particles):
            vc = []
            for i, _ in enumerate(model_particle_ids):
                con = scons[inverted_model_sort_mask[i]]
                vc.append(con)
            v.connections = vc
            ret = v
            break

    # Make sure all connections are in the vertex
    for c in cons:
        assert c in ret.connections
    # debug(f"{ret=}")
    if ret is None:
        raise Exception(
            f"Vertex {vertex} with cons {cons} not found in model\n{pdg_ids_list=}"
        )
    return ret



import re

from feynml.momentum import Momentum
from sympy.parsing.sympy_parser import parse_expr


def insert_mass(sname):
    """
    Handle strings of added and subtracted masses and place 'Mass' in front of them

    Example:
    >>> get_mass('0')
    0
    >>> get_mass('ZERO')
    0
    >>> get_mass('MC')
    MassMC
    """
    assert re.fullmatch(r"[a-zA-Z0-9+-_]+", sname), "Mass name does not match pattern"
    if sname == "0" or sname == "ZERO":
        return 0
    sname = re.sub(r"([a-zA-Z0-9_]+)", r"Mass_\1", sname)
    return sname


def insert_momentum(sname):
    """
    Handle strings of added and subtracted Momenta and place 'Mom' in front of them

    Example:
    >>> get_momentum('p1+p2-p3')
    'Momp1+Momp2-Momp3'
    >>> get_momentum('p1-l2')
    'Momp1-Moml2'
    >>> get_momentum('k1')
    'Momk1'
    >>> get_momentum('$k1$')
    'Momk1'
    >>> get_momentum('$p1$-$l2$')
    'Momp1-Moml2'
    """
    sname = sname.replace("$", "")
    assert re.fullmatch(
        r"[a-zA-Z0-9+-_]+", sname
    ), "Momentum name does not match pattern"
    sname = re.sub(r"([a-zA-Z0-9_]+)", r"Mom_\1", sname)
    return sname


def set_missing_momenta(feynman_diagram):
    fd = feynman_diagram
    for _, l in enumerate(fd.legs):
        if l.momentum is None:
            if l.is_incoming():
                l.with_momentum(Momentum(name="p" + str(l.id)))
            elif l.is_outgoing():
                l.with_momentum(Momentum(name="k" + str(l.id)))
    mv = fd.vertices
    for v in mv:
        cs = fd.get_connections(v)
        n_no_mom = 0
        no_mom = None
        for c in cs:
            if c.momentum is None or c.momentum.name is None:
                n_no_mom += 1
                no_mom = c
        if n_no_mom == 1:
            the_moms = ""
            for c in cs:
                if c.id != no_mom.id:
                    the_moms += "+" if c.goes_into(v) else "-"
                    the_moms += c.momentum.name
                no_mom.momentum = Momentum(name=the_moms)
    return fd



from feynamp.leg import get_leg_math
from feynamp.propagator import get_propagator_math
from feynamp.vertex import get_vertex_math


def get_math(obj, fd, model):
    if obj in fd.vertices:
        return get_vertex_math(obj, fd, model)
    elif obj in fd.legs:
        return get_leg_math(obj, fd, model)
    elif obj in fd.propagators:
        return get_propagator_math(obj, fd, model)
    else:
        raise Exception(f"Object {obj} not found in feynman diagram")



from typing import List

from feynml.feynmandiagram import FeynmanDiagram
from feynmodel.feyn_model import FeynModel

from feynamp.leg import get_leg_math_string
from feynamp.log import debug
from feynamp.propagator import get_propagator_math_string
from feynamp.vertex import get_vertex_math_string


def feynman_diagram_to_string(feynman_diagram, feyn_model):
    fd = feynman_diagram
    # TODO use these informations from qgraf, might need to be implemented in feynml
    # sign = fd.get_sign()
    # symmetry = fd.get_symmetry_factor()
    vm = []
    lm = []
    pm = []
    for v in fd.vertices:
        vm.append("(" + get_vertex_math_string(fd, v, feyn_model) + ")")
    for l in fd.legs:
        lm.append("(" + get_leg_math_string(fd, l, feyn_model) + ")")
    for p in fd.propagators:
        pm.append("(" + get_propagator_math_string(fd, p, feyn_model) + ")")

    ret = ""
    if len(vm) > 0:
        ret += f"{' * '.join(vm)} * "
    if len(lm) > 0:
        ret += f"{' * '.join(lm)} * "
    if len(pm) > 0:
        ret += f"{' * '.join(pm)} * "
    debug(f"{ret=}")
    return ret[0:-3]


def multiply(
    lst_fd1: List[FeynmanDiagram], lst_fd2: List[FeynmanDiagram], feyn_model: FeynModel
):
    # TODO should this care about fermion lines!?
    s = ""
    lst_fd1 = [feynman_diagram_to_string(l, feyn_model) for l in lst_fd1]
    lst_fd2 = [feynman_diagram_to_string(l, feyn_model) for l in lst_fd2]
    for fd1 in lst_fd1:
        for fd2 in lst_fd2:
            s += f"({fd1})*({fd2}) + "
    return s[:-3]


def square(lst_fd: List[FeynmanDiagram], feyn_model: FeynModel, tag=False):
    """
    Squares the list of feynman diagrams taking the fermion sign into account.
    """
    dims = lst_fd[0].get_externals_size()
    for fd in lst_fd:
        assert (
            dims == fd.get_externals_size()
        ), "All FeynmanDiagrams must have the same external legs"
    # TODO handle relative fermion sign (also majorana!) https://cds.cern.ch/record/238903/files/th-6549-92.pdf
    # TODO also multiply by the symmetry factor from qgraf
    # return multiply(lst_fd,[l.conjugated() for l in lst_fd],feyn_model)
    s = ""
    lst_fd1 = [feynman_diagram_to_string(l, feyn_model) for l in lst_fd]
    lst_fd2 = [feynman_diagram_to_string(l.conjugated(), feyn_model) for l in lst_fd]
    debug(f"{lst_fd1=}")
    # TODO this could also be done in multiply by comparing the diagrams
    for i in range(len(lst_fd1)):
        for j in range(i, len(lst_fd2)):
            sfd1 = lst_fd1[i]
            sfd2 = lst_fd2[j]
            if i == j:
                ttag = ""
                if tag:
                    ttag = f"*fd{lst_fd[i].id}*fd{lst_fd[i].id}fd{lst_fd[i].id}"
                s += f"({sfd1})*({sfd2}){ttag} + "
            elif i < j:
                ttag = ""
                if tag:
                    ttag = f"*fd{lst_fd[i].id}*fd{lst_fd[j].id}*fd{lst_fd[i].id}fd{lst_fd[j].id}"
                ferm_fac = lst_fd[i].get_fermion_factor(lst_fd[j])
                s += f"2*(+{sfd1})*({sfd2}){ttag}*{ferm_fac} + "  # TODO this needs Re!
    return s[:-3]


def add(lst_fd, feyn_model):
    lst_fd1 = [feynman_diagram_to_string(l, feyn_model) for l in lst_fd]
    ferm_facs = [lst_fd[0].get_fermion_factor(l) for l in lst_fd]
    s = ""
    for i in range(len(lst_fd1)):
        s += f"({lst_fd1[i]})*{ferm_facs[i]} + "
    return s[:-3]



from typing import List

from feynml.feynmandiagram import FeynmanDiagram
from feynml.leg import Leg
from feynmodel.feyn_model import FeynModel

from feynamp.form.form import get_dummy_index, init, run, string_to_form
from feynamp.leg import find_leg_in_model
from feynamp.log import debug
from feynamp.momentum import insert_momentum
from feynamp.util import is_mass_zero

gammas = """
repeat;
* identity
    id Gamma(Mua?,Spinb?,Spinc?) * GammaId(Spinc?,Spind?) = Gamma(Mua,Spinb,Spind);
    id Gamma(Mua?,Spinb?,Spinc?) * GammaId(Spind?,Spinb?) = Gamma(Mua,Spind,Spinc);
    id GammaId(Spina?,Spinb?) * GammaId(Spinb?,Spind?) = GammaId(Spina,Spind);
    id GammaId(Spina?,Spina?) = 4;
* Metric
    id Metric(Mua?,Mub?) * Gamma(Mua?,Spind?,Spinf?) = Gamma(Mub,Spind,Spinf);
    id Metric(Mua?,Mub?) * P(Mua?,Momd?) = P(Mub,Momd);
    id Metric(Mua?,Mub?) * Metric(Mub?,Mua?) = 4;
    id Metric(Mua?,Mub?) * Metric(Mub?,Muc?) = Metric(Mua,Muc);
    id Metric(Mua?,Mua?) = 4;
* standard Gamma algebra
    id Gamma(Mua?,Spinb?,Spinc?)*Gamma(Mua?,Spinc?,Spind?) = -GammaId(Spinb,Spind);
    id Gamma(Mua?,Spinb?,Spinc?)*Gamma(Mud?,Spinc?,Spine?)*Gamma(Mua?,Spine?,Spinf?) = -2*Gamma(Mud,Spinb,Spinf);
    id Gamma(Mua?,Spinb?,Spinc?)*Gamma(Mud?,Spinc?,Spine?)*Gamma(Muf?,Spine?,Spinm?)*Gamma(Mua?,Spinm?,Spink?) = 4*Metric(Mud,Muf)*GammaId(Spinb,Spink);
    id Gamma(Mua?,Spinb?,Spinc?)*Gamma(Mud?,Spinc?,Spine?)*Gamma(Muf?,Spine?,Spinm?)*Gamma(Muk?,Spinm?,Spinl?)*Gamma(Mua?,Spinl?,Spinj?) = -2*Gamma(Muk,Spinb,Spinc)*Gamma(Muf,Spinc,Spinl)*Gamma(Mud,Spinl,Spinj);
* traces of Gamma
    id Gamma(Mua?,Spinb?,Spinb?) = 0;
    id Gamma(Mua?,Spinb?,Spinc?)*Gamma(Mub?,Spinc?,Spinb?) = 4*Metric(Mua,Mub);
    id Gamma(Mua?,Spinb?,Spinc?)*Gamma(Mub?,Spinc?,Spind?)*Gamma(Muc?,Spind?,Spinb?) = 0;
    id Gamma(Mua?,Spinb?,Spinc?)*Gamma(Mub?,Spinc?,Spind?)*Gamma(Muc?,Spind?,Spine?)*Gamma(Mud?,Spine?,Spinb?) 
        = 4*(Metric(Mua,Mub)*Metric(Muc,Mud) - Metric(Mua,Muc)*Metric(Mub,Mud)+ Metric(Mua,Mud)*Metric(Mub,Muc)) ;
    id Gamma(Mua?,Spinb?,Spinc?)*Gamma(Mub?,Spinc?,Spind?)*Gamma(Muc?,Spind?,Spine?)*Gamma(Mud?,Spine?,Spinf?)*Gamma(Mue?,Spinf?,Sping?)*Gamma(Muf?,Sping?,Spinb?)
        = 4*(Metric(Mua,Mub)*Metric(Muc,Mud)*Metric(Mue,Muf) 
           - Metric(Mua,Mub)*Metric(Muc,Mue)*Metric(Mud,Muf) 
           + Metric(Mua,Mub)*Metric(Muc,Muf)*Metric(Mud,Mue) 
           - Metric(Mua,Muc)*Metric(Mub,Mud)*Metric(Mue,Muf) 
           + Metric(Mua,Muc)*Metric(Mub,Mue)*Metric(Mud,Muf) 
           - Metric(Mua,Muc)*Metric(Mub,Muf)*Metric(Mud,Mue) 
           + Metric(Mua,Mud)*Metric(Mub,Muc)*Metric(Mue,Muf) 
           - Metric(Mua,Mud)*Metric(Mub,Mue)*Metric(Muc,Muf) 
           + Metric(Mua,Mud)*Metric(Mub,Muf)*Metric(Muc,Mue)
           - Metric(Mua,Mue)*Metric(Mub,Muc)*Metric(Mud,Muf)
           + Metric(Mua,Mue)*Metric(Mub,Mud)*Metric(Muc,Muf)
           - Metric(Mua,Mue)*Metric(Mub,Muf)*Metric(Muc,Mud)
           + Metric(Mua,Muf)*Metric(Mub,Muc)*Metric(Mud,Mue)
           - Metric(Mua,Muf)*Metric(Mub,Mud)*Metric(Muc,Mue)
           + Metric(Mua,Muf)*Metric(Mub,Mue)*Metric(Muc,Mud));
endrepeat;
"""


# TODO implement collecting of gammas and form calc solving of it
# idea: GammaCollect(1,spin1,spin2, mu,nu,...) run through and then apply to expression
gamma_collect = """
repeat;
id Metric(Mua?,Mub?) = d_(Mua,Mub);
endrepeat;
#do i = 1, 10
once Gamma(Mux?,Spin1?,Spin2?) = GammaCollect(`i',Spin1,Spin2,Mux);
repeat;
id GammaCollect(`i',Spin1?,Spin2?,?mus)*Gamma(Mux?, Spin2?,Spin3?) = GammaCollect(`i',Spin1,Spin3, ?mus, Mux);
id GammaCollect(`i',Spin1?,Spin2?,?mus)*GammaId(Spin2?,Spin3?) = GammaCollect(`i',Spin1,Spin3, ?mus)*gi_(`i');
endrepeat;
id GammaCollect(`i',Spin1?,Spin1?,?mus) = g_(`i',?mus);
trace4, `i';
#enddo
#do i = 11, 21
once GammaId(Spin1?,Spin2?) = GammaIdCollect(`i',Spin1,Spin2);
repeat;
id GammaIdCollect(`i',Spin1?,Spin2?)*GammaId(Spin2?,Spin3?) = GammaIdCollect(`i',Spin1,Spin3)*gi_(`i');
endrepeat;
id GammaIdCollect(`i',Spin1?,Spin1?) = 1;
trace4, `i';
#enddo
"""


def get_gammas():
    return new_get_gammas()


def new_get_gammas():
    return get_dirac_trick() + gamma_collect


def get_gammas_v1():
    return get_polarisation_sum_v1() + get_dirac_trick() + gammas


def apply_gammas(string_expr):
    s = string_to_form(string_expr)
    return run(init + f"Local TMP = {s};" + get_gammas())


def get_orthogonal_polarisation_momentum(
    leg: Leg, fds: List[FeynmanDiagram], model: FeynModel
):
    for fd in [fds[0]]:
        for fleg in fd.legs:
            p = find_leg_in_model(fd, fleg, model)
            if is_mass_zero(p) and fleg != leg:
                mom = insert_momentum(fleg.momentum.name)
                return mom
    raise ValueError("No orthogonal momentum found")


def get_polarisation_sums(fds: List[FeynmanDiagram], model: FeynModel):
    pol_sums = ""
    # TODO might want to loop over all fds?
    for fd in [fds[0]]:
        for l in fd.legs:
            p = find_leg_in_model(fd, l, model)
            mom = insert_momentum(l.momentum.name)
            # mass = insert_mass(string_to_form(p.mass.name))
            if p.spin == 3:
                if is_mass_zero(p):
                    if p.color == 8:
                        mom_n = get_orthogonal_polarisation_momentum(l, fds, model)
                        pol_sums += get_polarisation_sum_physical(mom, mom_n)
                        # pol_sums += get_polarisation_sum_feynman(mom)
                        debug(f"mom: {mom}, mom_n: {mom_n}")
                    elif p.color == 1:
                        pol_sums += get_polarisation_sum_feynman(mom)
                else:
                    pol_sums += get_polarisation_sum_massive(mom)
    debug(f"pol_sums: {pol_sums}")
    return pol_sums


def get_polarisation_sum_massive(mom_a):
    pol_sum = f"""
    id epsstar(Muc?,Polb?,{mom_a}) * eps(Mul?,Pold?,{mom_a}) = -Metric(Mul,Muc) + (P(Mul,{mom_a})*P(Muc,{mom_a}))*Den({mom_a}.{mom_a});
    """
    return pol_sum


def get_polarisation_sum_feynman(mom_a):
    pol_sum = f"""
    id epsstar(Muc?,Polb?,{mom_a}) * eps(Mul?,Pold?,{mom_a}) = -Metric(Mul,Muc);
    """
    return pol_sum


def get_polarisation_sum_physical(mom_a, mom_b):
    pol_sum = f"""
    id epsstar(Muc?,Polb?,{mom_a}) * eps(Mul?,Pold?,{mom_a}) = -Metric(Muc,Mul) 
    + (P(Muc,{mom_a})*P(Mul,{mom_b}) +  P(Mul,{mom_a})*P(Muc,{mom_b}))*Den({mom_b}.{mom_a}) 
    - P(Muc,{mom_a})*P(Mul,{mom_a})*({mom_b}.{mom_b})*Den({mom_b}.{mom_a})*Den({mom_b}.{mom_a});
    """
    return pol_sum


def get_polarisation_sum(mom_a, mom_b=None):
    if mom_b is None:
        # massive case
        return get_polarisation_sum_massive(mom_a)
    elif mom_b == 0:
        # photon massless case
        return get_polarisation_sum_feynman(mom_a)
    else:
        # gluon massless case
        return get_polarisation_sum_physical(mom_a, mom_b)


def get_polarisation_sum_v1():
    polsum_feyn = """
    id epsstar(Muc?,Polb?,Moma?) * eps(Mul?,Pold?,Moma?) = -Metric(Muc,Mul);
    """
    polsum_feyn = """
    id epsstar(Muc?,Polb?,Moma?) * eps(Mul?,Pold?,Moma?) = -Metric(Mul,Muc);
    """
    return polsum_feyn


# TODO figure out why this does not work but manual dummies work
dirac_trick = """
repeat;
    id u(Spinc?,Momb?)*ubar(Spina?,Momb?) = Gamma(N5_?,Spinc,Spina) * P(N5_?,Momb) + GammaId(Spinc,Spina) * P(N5_?,Momb) * P(N5_?,Momb);
    id vbar(Spinc?,Momb?)*v(Spina?,Momb?) = Gamma(N7_?,Spinc,Spina) * P(N7_?,Momb) - GammaId(Spinc,Spina) * P(N7_?,Momb) * P(N7_?,Momb);
endrepeat;
"""


def get_dirac_trick(N=10):
    ret = ""
    for i in range(N):
        dummy = get_dummy_index()
        dirac_trick = f"""
    once u(Spinc?,Momb?)*ubar(Spina?,Momb?) = Gamma({dummy},Spinc,Spina) * P({dummy},Momb) + GammaId(Spinc,Spina) * P({dummy},Momb) * P({dummy},Momb);
    """
        ret += dirac_trick
    for i in range(N):
        dummy = get_dummy_index()
        dirac_trick = f"""
    once vbar(Spinc?,Momb?)*v(Spina?,Momb?) = Gamma({dummy},Spinc,Spina) * P({dummy},Momb) - GammaId(Spinc,Spina) * P({dummy},Momb) * P({dummy},Momb);
    """
        ret += dirac_trick
    return ret


# TODO: look above
# def get_dirac_trick():
#    return dirac_trick


def apply_dirac_trick(string_expr):
    s = string_to_form(string_expr)
    return run(init + f"Local TMP = {s};" + get_dirac_trick())


def apply_polarisation_sum(string_expr):
    s = string_to_form(string_expr)
    return run(init + f"Local TMP = {s};" + get_polarisation_sum())


# TODO: implement this use forms gamma algebra
def replace_indices_by_line():
    # return list of lines
    pass



from sympy.parsing.sympy_parser import parse_expr

from feynamp.sympy.color import identity
from feynamp.sympy.lorentz import gamma


def string_to_sympy(s):
    s = s.replace(
        "Gamma", "gamma"
    )  # we keep string like the ufo, but want lowercase gamma for sympy
    s = s.replace("Metric", "metric")
    s = s.replace("complex(0,1)", "I")  # sympy uses I for imaginary unit
    return parse_expr(s, local_dict={"gamma": gamma, "Identity": identity})
    # return parse_expr(s, evaluate=False)


def fancy_print(expr):
    s = string_to_sympy(expr)
    s.replace()



# Utilities for feynamp
from feynmodel.particle import Particle


def safe_index_replace(string, old, new):
    string = string.replace("," + old + ",", "," + new + ",")
    string = string.replace("(" + old + ",", "(" + new + ",")
    string = string.replace("," + old + ")", "," + new + ")")
    return string


def find_particle_in_model(particle, model):
    for pp in model.particles:
        if pp.pdg_code == particle.pdgid:
            pp.particle = particle
            return pp
    return None


def is_mass_zero(p: Particle):
    return p.mass.name == "ZERO" or float(p.mass.value) == 0.0



from feynamp.form.form import init, run, string_to_form

color = """
**********************************************************
*                  COLOUR STRUCTURE SIMPLIFY             *
**********************************************************
    
repeat;
* remove df(k,j)
   id df(k?,l?)*df(l?,j?)=df(k,j);
   id T(a?,k?,l?)*df(k?,j?)=T(a,j,l);
   id T(a?,k?,l?)*df(l?,j?)=T(a,k,j);
* remove da(a,b)
   id da(a?,b?)*da(b?,c?)=da(a,c);
   id T(a?,k?,l?)*da(a?,b?)=T(b,k,l);
   id f(a?,b?,c?)*da(a?,d?)=f(d,b,c);
* length-three objects simplify:
   id T(b?,k?,j?)*T(a?,j?,c?)*T(b?,c?,l?)=(-Tr/Nc*T(a,k,l));
   id T(b?,j?,l?)*T(c?,l?,k?)*f(a?,b?,c?)=(i_*Nc*Tr*T(a,j,k));
* length-two objects that give out df(k,j)
   id T(a?,c?,j?)*T(a?,k?,l?)=(-1/Nc*df(c, j)*df(k, l)*Tr + df(c, l)*df(j, k)*Tr);
   id T(a?,k?,l?)*T(a?,l?,j?)=(Cf*df(k,j));
* length-two objects that give out da(a,b)
   id T(a?,k?,l?)*T(b?,l?,k?)=(Tr*da(a,b));
   id f(a?,b?,c?)*f(d?,b?,c?)=Nc*da(a,d); 
* simplify traces
   id T(b?,k?,k?)=0;
   id da(a?,a?)=Nc*Cf/Tr;
   id df(a?,a?)=Nc;
* simplify combination of factors
*   id Nc^-2=2-Nc^2+Cf^2*Tr^-2;
*   id Nc^2=1+Nc*Cf/Tr;
   id Tr=1/2;
   id Tr^-1=2;
* double f(a,b,c) simplify
*  id f(a?,b?,e?)*f(c?,d?,e?)=-2 * { [T(a), T(b)] [T(c), T(d)]};
*   id f(a?,b?,e?)*f(c?,d?,e?)=-2 * (T(a,e,N1_?)*T(b,N1_?,N2_?) - T(b,e,N1_?)*T(a,N1_?,N2_?))*(T(c,N2_?,N3_?)*T(d,N3_?,e)-T(d,N2_?,N3_?)*T(c,N3_?,e));
endrepeat;
"""

color_sum = """
**********************************************************
*                  COLOUR SUM SIMPLIFY                   *
**********************************************************
repeat;
  id VA(Glua?,Momb?)*VA(Gluc?,Momb?) = da(Glua,Gluc);
  id VC(Cola?,Momb?)*VC(Colc?,Momb?) = df(Cola,Colc);
endrepeat;
"""


def get_color():
    return color_sum + color


def apply_color(string_expr):
    s = string_to_form(string_expr)
    return run(init + f"Local TMP = {s};" + get_color())


def apply_color_sum(string_expr):
    s = string_to_form(string_expr)
    return run(init + f"Local TMP = {s};" + color_sum)


def apply_color_simplify(string_expr):
    s = string_to_form(string_expr)
    return run(init + f"Local TMP = {s};" + color)



import os
import re
from typing import List

import form
import sympy
from feynml.feynmandiagram import FeynmanDiagram
from feynmodel.feyn_model import FeynModel

import feynamp.amplitude as amplitude
from feynamp import get_color_average, get_spin_average
from feynamp.form.color import get_color
from feynamp.form.lorentz import get_gammas, get_polarisation_sums
from feynamp.form.momentum import (
    apply,
    apply_den,
    get_kinematics,
    get_mandelstamm,
    get_onshell,
)
from feynamp.log import debug

# TODO compute squared  functino which coutns legs!!"!!!" and picks right mandelstamm,s


def compute_squared(fds: List[FeynmanDiagram], fm: FeynModel, tag=False):
    dims = fds[0].get_externals_size()
    for fd in fds:
        assert (
            dims == fd.get_externals_size()
        ), "All FeynmanDiagrams must have the same external legs"
    s2 = amplitude.square(fds, fm, tag=tag)
    debug(f"{s2=}")
    fs = ""
    fs += get_polarisation_sums(fds, fm)
    fs += get_gammas()
    fs += get_color()
    fs += get_kinematics()
    fs += get_onshell(fds, fm)
    fs += get_mandelstamm(fds, fm)

    rs = apply(s2, fs)
    debug(f"{rs=}")

    rr = apply_den(
        rs,
        get_onshell(fds, fm) + get_mandelstamm(fds, fm),
    )
    debug(f"{rr=}")

    ret = sympy.simplify(
        sympy.parse_expr(
            rr.replace("Mom_", "")
            .replace(".", "_")
            .replace("^", "**")
            .replace("mss", "s")
            .replace("msu", "u")
            .replace("mst", "t")
        )
        * sympy.parse_expr("*".join([*get_color_average(fds), *get_spin_average(fds)]))
    )
    return ret



import re
from typing import List

from feynml.feynmandiagram import FeynmanDiagram
from feynmodel.feyn_model import FeynModel

# from feynamp.form import *
from feynamp.form.form import init, run, string_to_form
from feynamp.leg import find_leg_in_model
from feynamp.log import warning
from feynamp.momentum import insert_mass, insert_momentum

momenta = """
repeat;
    id P(Mu1?,Moma?)*P(Mu1?,Momb?) = Moma.Momb;
endrepeat;
"""


def get_kinematics():
    return get_momenta() + get_denominators()


def apply_kinematics(string_expr):
    s = string_to_form(string_expr)
    return run(init + f"Local TMP = {s};" + get_kinematics())


def get_momenta():
    return momenta


def apply_momenta(string_expr):
    s = string_to_form(string_expr)
    return run(init + f"Local TMP = {s};" + momenta)


denominators = """
repeat;
    id Denom(Mom1?,Massa?) = Den(Mom1.Mom1-Massa^2);
endrepeat;
"""


def get_denominators():
    return denominators


def apply_denominators(string_expr):
    s = string_to_form(string_expr)
    return run(init + f"Local TMP = {s};" + denominators)


def apply(string_expr, str_a):
    s = string_to_form(string_expr)
    return run(init + f"Local TMP = {s};" + str_a)


def apply_den(string_expr, str_f):
    # re match all Dens
    s = string_expr
    res = re.findall(r"Den\(([a-zA-Z0-9_+*-\.^]+)\)", string_expr)
    if res:
        for og in res:  # TODO parallelize? each as one var in form?
            g = apply(og, str_f)
            s = s.replace("Den(" + og + ")", "1/(" + g + ")")
    return s


def get_onshell(fds: List[FeynmanDiagram], model: FeynModel):
    if isinstance(fds, FeynmanDiagram):
        warning(
            "get_onshell is deprecated, use get_onshell with list of FeynmanDiagram"
        )
        fds = [fds]
    r = ""
    # TODO might be redundant
    for fd in fds:
        for l in fd.legs:
            p = find_leg_in_model(fd, l, model)
            mom = insert_momentum(l.momentum.name)
            mass = insert_mass(string_to_form(p.mass.name))
            r += f"id {mom}.{mom} = {mass}^2;\n"
    return r


def apply_onshell(string_expr, fd, model):
    s = string_to_form(string_expr)
    return run(init + f"Local TMP = {s};" + get_onshell(fd, model))


def get_mandelstamm(fds: List[FeynmanDiagram], model: FeynModel, *args, **kwargs):
    if fds[0].get_externals_size() == (2, 2):
        return get_mandelstamm_2_to_2(fds, model, *args, **kwargs)
    elif fds[0].get_externals_size() == (2, 3):
        return get_mandelstamm_2_to_3(fds, model, *args, **kwargs)
    else:
        raise ValueError("Only 2 to 2 and 2 to 3 Mandelstamm are supported")


def get_mandelstamm_2_to_2(
    fds: List[FeynmanDiagram],
    model: FeynModel,
    replace_s=False,
    replace_t=False,
    replace_u=False,
):
    if isinstance(fds, FeynmanDiagram):
        warning(
            "get_onshell is deprecated, use get_onshell with list of FeynmanDiagram"
        )
        fds = [fds]
    r = ""
    for fd in fds:
        li = []
        lo = []
        for f in fd.legs:
            if f.is_incoming():
                li.append(f)
            elif f.is_outgoing():
                lo.append(f)
            else:
                raise ValueError("Leg is neither incoming nor outgoing")
        l1, l2 = li
        l3, l4 = lo
        p1 = find_leg_in_model(fd, l1, model)
        mom1 = insert_momentum(l1.momentum.name)
        mass1 = insert_mass(string_to_form(p1.mass.name))
        p2 = find_leg_in_model(fd, l2, model)
        mom2 = insert_momentum(l2.momentum.name)
        mass2 = insert_mass(string_to_form(p2.mass.name))
        p3 = find_leg_in_model(fd, l3, model)
        mom3 = insert_momentum(l3.momentum.name)
        mass3 = insert_mass(string_to_form(p3.mass.name))
        p4 = find_leg_in_model(fd, l4, model)
        mom4 = insert_momentum(l4.momentum.name)
        mass4 = insert_mass(string_to_form(p4.mass.name))
        r += f"id {mom1}.{mom2} = mss/2-{mass1}^2/2-{mass2}^2/2;\n"
        r += f"id {mom3}.{mom4} = mss/2-{mass3}^2/2-{mass4}^2/2;\n"
        r += f"id {mom1}.{mom3} = -mst/2+{mass1}^2/2+{mass3}^2/2;\n"
        r += f"id {mom4}.{mom2} = -mst/2+{mass4}^2/2+{mass2}^2/2;\n"
        r += f"id {mom1}.{mom4} = -msu/2+{mass1}^2/2+{mass4}^2/2;\n"
        r += f"id {mom2}.{mom3} = -msu/2+{mass2}^2/2+{mass3}^2/2;\n"
        if replace_s:
            r += f"id mss = -msu-mst+{mass2}^2+{mass3}^2+{mass4}^2+{mass1}^2;\n"
        if replace_t:
            r += f"id mst = -mss-msu+{mass2}^2+{mass3}^2+{mass4}^2+{mass1}^2;\n"
        if replace_u:
            r += f"id msu = -mss-mst+{mass2}^2+{mass3}^2+{mass4}^2+{mass1}^2;\n"
    return r


def get_mandelstamm_2_to_3(
    fds: List[FeynmanDiagram],
    model: FeynModel
    # , replace_s=False, replace_t=False, replace_u=False
):
    if isinstance(fds, FeynmanDiagram):
        warning(
            "get_onshell is deprecated, use get_onshell with list of FeynmanDiagram"
        )
        fds = [fds]
    r = ""

    for fd in fds:
        li = []
        lo = []
        for f in fd.legs:
            if f.is_incoming():
                li.append(f)
            elif f.is_outgoing():
                lo.append(f)
            else:
                raise ValueError("Leg is neither incoming nor outgoing")
        l1, l2 = li
        l3, l4, l5 = lo
        p1 = find_leg_in_model(fd, l1, model)
        mom1 = insert_momentum(l1.momentum.name)
        mass1 = insert_mass(string_to_form(p1.mass.name))
        p2 = find_leg_in_model(fd, l2, model)
        mom2 = insert_momentum(l2.momentum.name)
        mass2 = insert_mass(string_to_form(p2.mass.name))
        p3 = find_leg_in_model(fd, l3, model)
        mom3 = insert_momentum(l3.momentum.name)
        mass3 = insert_mass(string_to_form(p3.mass.name))
        p4 = find_leg_in_model(fd, l4, model)
        mom4 = insert_momentum(l4.momentum.name)
        mass4 = insert_mass(string_to_form(p4.mass.name))
        p5 = find_leg_in_model(fd, l5, model)
        mom5 = insert_momentum(l5.momentum.name)
        mass5 = insert_mass(string_to_form(p5.mass.name))

        # r += f"id {mom5} = {mom1} + {mom2} - {mom3} - {mom4};\n"
        r += f"id {mom1}.{mom2} = mss12/2-{mass1}^2/2-{mass2}^2/2;\n"
        r += f"id {mom3}.{mom4} = mss34/2-{mass3}^2/2-{mass4}^2/2;\n"
        r += f"id {mom3}.{mom5} = mss35/2-{mass3}^2/2-{mass5}^2/2;\n"
        r += f"id {mom4}.{mom5} = mss45/2-{mass4}^2/2-{mass5}^2/2;\n"

        r += f"id {mom1}.{mom3} = -mst13/2+{mass1}^2/2+{mass3}^2/2;\n"
        r += f"id {mom1}.{mom4} = -mst14/2+{mass1}^2/2+{mass4}^2/2;\n"
        r += f"id {mom1}.{mom5} = -mst15/2+{mass1}^2/2+{mass5}^2/2;\n"

        r += f"id {mom2}.{mom3} = -mst23/2+{mass2}^2/2+{mass3}^2/2;\n"
        r += f"id {mom2}.{mom4} = -mst24/2+{mass2}^2/2+{mass4}^2/2;\n"
        r += f"id {mom2}.{mom5} = -mst25/2+{mass2}^2/2+{mass5}^2/2;\n"

    return r


def apply_mandelstamm_2_to_2(
    string_expr,
    fd,
    model,
):
    s = string_to_form(string_expr)
    return run(init + f"Local TMP = {s};" + get_mandelstamm_2_to_2(fd, model))



from re import I

from sympy import Dummy, Function, I, Wild, cacheit, symbols

eps = Function("eps")
eps_star = Function("eps_star")
u = Function("u")
u_bar = Function("u_bar")
v = Function("v")
v_bar = Function("v_bar")

gamma = Function("gamma")
gamma_id = Function("gamma_id")
ProjP = Function("P_+")
ProjM = Function("P_-")
metric = Function("eta")

P = Function("P")


def apply_polarisation_sum(expr):
    wmu, wnu, wc, wd, wi, wj, wk, ww = symbols("wmu wnu wc wd wi wj wk ww", cls=Wild)
    expr = expr.replace(
        eps_star(wmu, wd, wc) * eps(wnu, wi, wc), -metric(wmu, wnu), map=True
    )
    # TODO do the pol sums
    return expr


def apply_dirac_trick(expr):
    wa, wb, wc, wd, wi, wj, wk, ww = symbols("wa wb wc wd wi wj wk ww", cls=Wild)
    maps = True
    while maps:
        di, dj = symbols("di dj", cls=Dummy)
        expr, maps = expr.replace(
            u(wb, wa) * u_bar(wc, wa) * ww,
            gamma(di, wb, wc) * P(di, wa) * ww
            + gamma_id(wb, wc) * P(di, wa) * P(di, wa) * ww,
            map=True,
        )
    maps = True
    while maps:
        di, dj = symbols("di dj", cls=Dummy)
        expr, maps = expr.replace(
            v(wb, wa) * v_bar(wc, wa) * ww,
            gamma(dj, wb, wc) * P(dj, wa) * ww
            - gamma_id(wb, wc) * P(dj, wa) * P(dj, wa) * ww,
            map=True,
        )
    return expr


def apply_gammas(expr):
    expr = apply_5_gamma(expr)
    expr = apply_4_gamma(expr)
    expr = apply_3_gamma(expr)
    expr = apply_2_gamma(expr)
    return expr


def apply_2_gamma(expr):
    wa, wb, wc, wd, wi, wj, wk, ww = symbols("wa wb wc wd wi wj wk ww", cls=Wild)
    maps = True
    while maps:
        di, dj = symbols("di dj", cls=Dummy)
        expr, maps = expr.replace(
            gamma(wa, wb, wd) * gamma(wa, wd, wc) * ww, -gamma_id(wb, wc) * ww, map=True
        )
    return expr


def apply_3_gamma(expr):
    wa, wb, wc, wd, we, wg, wi, wj, wk, ww = symbols(
        "wa wb wc wd we wg wi wj wk ww", cls=Wild
    )
    maps = True
    while maps:
        di, dj = symbols("di dj", cls=Dummy)
        expr, maps = expr.replace(
            gamma(wa, wb, wd) * gamma(wg, wd, wc) * gamma(wa, wc, we) * ww,
            -2 * gamma(wg, wb, we) * ww,
            map=True,
        )
    return expr


def apply_4_gamma(expr):
    wa, wb, wc, wd, we, wf, wg, wh, wi, wj, wk, ww = symbols(
        "wa wb wc wd we wf wg wh wi wj wk ww", cls=Wild
    )
    maps = True
    while maps:
        di, dj = symbols("di dj", cls=Dummy)
        expr, maps = expr.replace(
            gamma(wa, wb, wd)
            * gamma(wi, wd, we)
            * gamma(wj, we, wf)
            * gamma(wa, wf, wc)
            * ww,
            4 * metric(wi, wj) * gamma_id(wb, wc) * ww,
            map=True,
        )
    return expr


def apply_5_gamma(expr):
    wa, wb, wc, wd, we, wf, wg, wh, wi, wj, wk, ww = symbols(
        "wa wb wc wd we wf wg wh wi wj wk ww", cls=Wild
    )
    maps = True
    while maps:
        print(maps)
        di, dj = symbols("di dj", cls=Dummy)
        expr, maps = expr.replace(
            gamma(wa, wb, wd)
            * gamma(wi, wd, we)
            * gamma(wj, we, wf)
            * gamma(wk, wf, wg)
            * gamma(wa, wg, wh)
            * ww,
            -2 * gamma(wi, wb, di) * gamma(wj, di, dj) * gamma(wk, dj, wh) * ww,
            map=True,
        )
    return expr



import os
import re

import form

count = 0
dummy = 0
init = """
Symbols Pi,G,ZERO,Tr,Nc,Cf,CA,MC,ee;
AutoDeclare Index Mu,Spin,Pol,Col,Glu,Propagator;
AutoDeclare Symbol Mass,fd,mss,mst,msu;
AutoDeclare Vector Mom;
Tensors f(antisymmetric),Metric(symmetric),df(symmetric),da(symmetric),Identity(symmetric);
Function ProjM,ProjP,VF,xg,xgi,P,dg,dgi,xeg,xegi;
CFunctions Den,T,Denom,P,Gamma,u,v,ubar,vbar,eps,epsstar,VC,VA,GammaId, GammaCollect, GammaIdCollect;
Indices a,o,n,m,tm,tn,beta,b,m,betap,alphap,a,alpha,ind,delta,k,j,l,c,d,e;
"""


def get_dummy_index():
    global dummy
    dummy = dummy + 1
    return f"N{dummy}_?"


def string_to_form(s):
    s = s.replace("complex(0,1)", "i_")  # form uses i_ for imaginary unit
    s = s.replace("Gamma_Id", "GammaId")
    s = s.replace("u_bar", "ubar")
    s = s.replace("v_bar", "vbar")
    s = s.replace("eps_star", "epsstar")
    s = s.replace("Identity", "df")  # TODO check if this holds or also happens for anti
    s = s.replace("ZERO", "0")
    s = s.replace(".*", "*")  # handle decimals
    s = s.replace(".)", ")")  # handle decimals
    return s


def run(s, show=False, keep_form_file=True):
    global count
    count = count + 1
    with open("form" + str(count) + ".frm", "w") as frm:
        with form.open(keep_log=1000) as f:
            local = s.split("Local")[1].split("=")[0].strip()
            txt = s + "print " + local + ";.sort;"
            f.write(txt)
            frm.write(txt)
            r = f.read("" + local)
            r = re.sub(r"\+factor_\^?[0-9]*", r"", r).strip("*")
            if show:
                print(r + "\n")
            # delete frm file
            if not keep_form_file:
                os.remove("form" + str(count) + ".frm")
            return r


def sympyfy(string_expr):
    from sympy import simplify
    from sympy.parsing.sympy_parser import parse_expr

    ret = simplify(
        parse_expr(
            string_expr.replace("Mom_", "")
            .replace(".", "_")
            .replace("^", "**")
            .replace("mss", "s")
            .replace("msu", "u")
            .replace("mst", "t")
        )
    )
    return simplify(ret.subs("Nc", "3").subs("Cf", "4/3"))



from re import I

from sympy import Dummy, Function, I, Wild, cacheit, symbols

# from symengine import *


N_c = symbols("N_c", integer=True)
N_g = symbols("N_g", integer=True)
C_A = symbols("C_A", integer=True)
C_F = symbols("C_F", integer=True)
T_F = symbols("T_F", integer=True)

T = Function("T")
f = Function("f")

delta_c = Function("delta_c")
delta_g = Function("delta_g")
identity = Function("Identity")

VC = Function("VC")
VA = Function("VA")


def apply_color(expr, expand=True):
    # if expr is tuple
    if isinstance(expr, tuple):
        expr = expr[0]
    expr = apply_color_sum(expr)
    expr = apply_id(expr)
    expr = apply_f(expr)
    expr = apply_TT(expr)
    expr = apply_T(expr)
    expr = apply_id(expr)
    if expand:
        expr = expr.expand()  # so that we hit it
    expr = apply_delta_delta(expr)
    expr = apply_T_delta(expr)
    expr = apply_f_delta(expr)
    expr = apply_T(expr)
    expr = apply_id(expr)
    return expr


def apply_color_sum(expr):
    wi, wj, wk, ww = symbols("wi wj wk ww", cls=Wild)
    expr = expr.replace(VC(wi, wk) * VC(wj, wk) * ww, ww * delta_c(wi, wj))
    expr = expr.replace(VA(wi, wk) * VA(wj, wk) * ww, ww * delta_g(wi, wj))
    return expr


def apply_delta(expr):
    expr = apply_delta_delta(expr)
    expr = apply_T_delta(expr)
    expr = apply_f_delta(expr)

    return expr  # , bool(chagned1 or chagned2 or chagned3)


def apply_id(expr):
    wi = Wild("wi")
    expr, map1 = expr.replace(delta_c(wi, wi), N_c, map=True)
    expr, map2 = expr.replace(delta_g(wi, wi), N_g, map=True)
    return expr  # , bool(map1 or map2)


def apply_delta_delta(expr):
    wa, wb, wc, wd, wi, wj, wk, ww = symbols("wa wb wc wd wi wj wk ww", cls=Wild)
    expr, map1 = expr.replace(
        delta_c(wa, wb) * delta_c(wb, wc) * ww, delta_c(wa, wc) * ww, map=True
    )
    expr, map2 = expr.replace(
        delta_c(wa, wb) * delta_c(wc, wb) * ww, delta_c(wa, wc) * ww, map=True
    )
    expr, map3 = expr.replace(
        delta_c(wb, wa) * delta_c(wb, wc) * ww, delta_c(wa, wc) * ww, map=True
    )
    expr, map4 = expr.replace(
        delta_c(wb, wa) * delta_c(wc, wb) * ww, delta_c(wa, wc) * ww, map=True
    )
    return expr  # , bool(map1 or map2 or map3 or map4)


def apply_T_delta(expr):
    wa, wb, wc, wd, wi, wj, wk, ww = symbols("wa wb wc wd wi wj wk ww", cls=Wild)
    (
        d1,
        d2,
        d3,
        d4,
        d5,
        d6,
        d7,
        d8,
    ) = symbols("d1 d2 d3 d4 d5 d6 d7 d8", cls=Dummy)
    expr, map1 = expr.replace(
        ww * T(wa, wj, wi) * delta_c(wi, wk), ww * T(wa, wj, wk), map=True
    )
    expr, map2 = expr.replace(
        ww * T(wa, wi, wj) * delta_c(wi, wk), ww * T(wa, wk, wj), map=True
    )
    expr, map3 = expr.replace(
        ww * T(wa, wi, wj) * delta_c(wk, wi), ww * T(wa, wk, wj), map=True
    )
    expr, map4 = expr.replace(
        ww * T(wa, wj, wi) * delta_c(wk, wi), ww * T(wa, wj, wk), map=True
    )

    expr, map5 = expr.replace(
        ww * T(wa, wi, wj) * delta_g(wa, wb), ww * T(wb, wi, wj), map=True
    )
    expr, map6 = expr.replace(
        ww * T(wa, wi, wj) * delta_g(wb, wa), ww * T(wb, wi, wj), map=True
    )

    # expr = expr.replace(ww*T(wa,wi,wj)* delta_c(wi,wj), ww*T(wa,d1,d1))
    # expr = expr.replace(ww*T(wa,wi,wj)* delta_c(wj,wi), ww*T(wa,d2,d2))
    return expr  # , bool(map1 or map2 or map3 or map4 or map5 or map6)


def apply_f_delta(expr):
    wa, wb, wc, wd, wi, wj, wk, ww = symbols("wa wb wc wd wi wj wk ww", cls=Wild)
    (
        d1,
        d2,
        d3,
        d4,
        d5,
        d6,
        d7,
        d8,
    ) = symbols("d1 d2 d3 d4 d5 d6 d7 d8", cls=Dummy)
    expr, map1 = expr.replace(
        ww * f(wa, wb, wc) * delta_g(wa, wd), ww * f(wd, wb, wc), map=True
    )
    expr, map2 = expr.replace(
        ww * f(wa, wb, wc) * delta_g(wb, wd), ww * f(wa, wd, wc), map=True
    )
    expr, map3 = expr.replace(
        ww * f(wa, wb, wc) * delta_g(wc, wd), ww * f(wa, wb, wd), map=True
    )

    expr, map4 = expr.replace(
        ww * f(wa, wb, wc) * delta_g(wd, wa), ww * f(wd, wb, wc), map=True
    )
    expr, map5 = expr.replace(
        ww * f(wa, wb, wc) * delta_g(wd, wb), ww * f(wa, wd, wc), map=True
    )
    expr, map6 = expr.replace(
        ww * f(wa, wb, wc) * delta_g(wd, wc), ww * f(wa, wb, wd), map=True
    )

    expr, map7 = expr.replace(
        ww * f(wa, wb, wc) * delta_c(wa, wb), ww * f(d3, d3, wc), map=True
    )
    expr, map8 = expr.replace(
        ww * f(wa, wb, wc) * delta_c(wa, wc), ww * f(d4, wb, d4), map=True
    )
    expr, map9 = expr.replace(
        ww * f(wa, wb, wc) * delta_c(wb, wc), ww * f(wa, d5, d5), map=True
    )

    expr, map10 = expr.replace(
        ww * f(wa, wb, wc) * delta_c(wb, wa), ww * f(d6, d6, wc), map=True
    )
    expr, map11 = expr.replace(
        ww * f(wa, wb, wc) * delta_c(wc, wa), ww * f(d7, wb, d7), map=True
    )
    expr, map12 = expr.replace(
        ww * f(wa, wb, wc) * delta_c(wc, wb), ww * f(wa, d8, d8), map=True
    )
    return expr  # , bool(map1 or map2 or map3 or map4 or map5 or map6 or map7 or map8 or map9 or map10 or map11 or map12)


def apply_T(expr):
    wa, wi = symbols("wa wi", cls=Wild)
    # traceless
    expr, map = expr.replace(T(wa, wi, wi), 0, map=True)
    return expr  # , bool(map)


def apply_f(expr):
    wa, wb, wc, ww = symbols("wa wb wc ww", cls=Wild)
    i, j, k = symbols("i j k", cls=Dummy)
    expr, map = expr.replace(
        f(wa, wb, wc) * ww,
        2
        * I
        * (
            T(wc, i, j) * T(wb, j, k) * T(wa, k, i)
            - T(wa, i, j) * T(wb, j, k) * T(wc, k, i)
        )
        * ww,
        map=True,
    )
    return expr  # , bool(map)


def apply_TT(expr):
    """

    Examples
    ========

    >>> from feynpy.color import *
    >>> i,j,k,l,g = symbols('i j k l g')
    >>> m = T(g,i,j)*T(g,k,l)
    >>> apply_TT(m)
    >>> m
    -1/C_A*Color.delta(i, k)*Color.delta(j, l)/2 + Color.delta(i, l)*Color.delta(j, k)/2

    """
    wi, wj, wk, wl, wg, ww = symbols("wi wj wk wl wg ww", cls=Wild)
    expr, map = expr.replace(
        T(wg, wi, wj) * T(wg, wk, wl) * ww,
        (
            delta_c(wi, wl) * delta_c(wk, wj) / 2
            - 1 / N_c / 2 * delta_c(wi, wj) * delta_c(wk, wl)
        )
        * ww,
        map=True,
    )
    return expr  # , bool(map)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['feynamp', 'feynamp.form', 'feynamp.sympy']

package_data = \
{'': ['*']}

install_requires = \
['feynml>=0.2.26', 'feynmodel>=0.0.5', 'python-form', 'sympy']

setup_kwargs = {
    'name': 'feynamp',
    'version': '0.0.5.12',
    'description': 'Compute Feynman diagrams',
    'long_description': '# FeynAmp\n\n[![PyPI version][pypi image]][pypi link] [![PyPI version][pypi versions]][pypi link]  ![downloads](https://img.shields.io/pypi/dm/feynamp.svg)\n\n[![test][a t image]][a t link]      [![Coverage Status][c t i]][c t l]  [![Codacy Badge](https://app.codacy.com/project/badge/Coverage/b7192679265b441cb534c9dc06d1b350)](https://app.codacy.com/gh/APN-Pucky/feynamp/dashboard?utm_source=gh&utm_medium=referral&utm_content=&utm_campaign=Badge_coverage)   [![Codacy Badge](https://app.codacy.com/project/badge/Grade/b7192679265b441cb534c9dc06d1b350)](https://app.codacy.com/gh/APN-Pucky/feynamp/dashboard?utm_source=gh&utm_medium=referral&utm_content=&utm_campaign=Badge_grade)\n\n\n## Related\n\n* Mathematica: FormCalc, FeynCalc\n* Julia: https://arxiv.org/pdf/2310.07634.pdf\n\n[pypi image]: https://badge.fury.io/py/feynamp.svg\n[pypi link]: https://pypi.org/project/feynamp/\n[pypi versions]: https://img.shields.io/pypi/pyversions/feynamp.svg\n\n[a t link]: https://github.com/APN-Pucky/feynamp/actions/workflows/test.yml\n[a t image]: https://github.com/APN-Pucky/feynamp/actions/workflows/test.yml/badge.svg\n\n[c t l]: https://coveralls.io/github/APN-Pucky/feynamp?branch=master\n[c t i]: https://coveralls.io/repos/github/APN-Pucky/feynamp/badge.svg?branch=master\n',
    'author': 'Alexander Puck Neuwirth',
    'author_email': 'alexander@neuwirth-informatik.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/APN-Pucky/feynamp',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from feynml.id import generate_new_id

from feynamp.momentum import insert_mass, insert_momentum
from feynamp.util import find_particle_in_model


def get_propagator_math_string(fd, prop, model):
    return get_propagator_math(fd, prop, model)


def get_propagator_math(fd, prop, model):
    # find the particle in the model
    p = find_propagator_in_model(fd, prop, model)
    if p.particle.momentum is None or p.particle.momentum.name is None:
        raise ValueError("Momentum not set for particle")
    mom = insert_momentum(p.particle.momentum.name)
    mass = insert_mass(p.mass.name)
    ret = ""
    if p.color == 3 or p.color == -3:
        ret += f"df(ColIn{p.particle.id},ColOut{p.particle.id})*"
    if p.color == 8:
        ret += f"da(GluIn{p.particle.id},GluOut{p.particle.id})*"
    # if boson just 1/(p^2-m^2)
    if p.spin == 3:
        # nid = generate_new_id()
        # TODO treate denominators differently for loops etc?
        ret += f"Metric(MuIn{p.particle.id},MuOut{p.particle.id})*"
    elif p.spin == 2:  # TODO handle plus minus mass for fermions
        nid = generate_new_id()
        ret += f"(P(Mu{nid},{mom})*Gamma(Mu{nid},SpinIn{p.particle.id},SpinOut{p.particle.id})"
        ret += f" + {mass}*GammaId(SpinIn{p.particle.id},SpinOut{p.particle.id}))*"
    else:
        raise ValueError("Spin not set for particle")
    return ret + f"Denom({mom},{mass})"


def find_propagator_in_model(fd, prop, model):
    assert prop in fd.propagators
    return find_particle_in_model(prop, model)



from .amplitude import feynman_diagram_to_string
from .log import debug
from .propagator import find_propagator_in_model, get_propagator_math
from .vertex import find_vertex_in_model, get_vertex_math


def get_spin_average(fds):
    """
    Initial fermions produce an averaging by 1/2 each
    """
    r = []
    for leg in fds[0].get_incoming():
        if leg.is_any_fermion():
            r += ["1/2"]
        elif leg.pdgid == 21:
            r += ["1/2"]
        elif leg.pdgid == 22:
            r += ["1/2"]
        else:
            raise ValueError(f"TODO: Unknown color average for pdgid {leg.pdgid}")
    debug(f"get_spin_average():{r}")
    return r


def get_color_average(fds):
    """
    Initial quarks produce an averaging by 1/3 each
    Initial gluons produce an averaging by 1/8 each
    """
    r = []
    for leg in fds[0].get_incoming():
        if leg.pdgid == 21:
            r += ["1/8"]
        elif leg.pdgid in range(1, 7) or -leg.pdgid in range(1, 17):
            r += ["1/3"]
        elif leg.pdgid == 22:
            r += ["1"]
        elif leg.pdgid == 23:
            r += ["1"]
        elif leg.pdgid in range(11, 19) or -leg.pdgid in range(11, 19):
            r += ["1"]
        else:
            raise ValueError(f"TODO: Unknown color average for pdgid {leg.pdgid}")
    debug(f"get_color_average():{r}")
    return r
